"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
const { RedisQueueClient } = require("redis-ordered-queue");
const Redis = require("ioredis");
const { updateSortFieldsForDocument, generateSortIdForAllDocuments } = require("./utils");
let modelsQueue;
class ModelsQueue {
    constructor(redisOptions) {
        this.modelNameToModelMap = {};
        const redis = new Redis(redisOptions);
        this.client = new RedisQueueClient({
            redis,
            batchSize: 1,
            groupVisibilityTimeoutMs: 60000,
            pollingTimeoutMs: 10000,
            consumerCount: 1,
            redisKeyPrefix: "mongoose-sort-encrypted-field",
        });
        this.client.startConsumers({ handleMessage: this.handleMessage });
    }
    handleMessage({ data, context: { lock: { groupId }, }, }) {
        return __awaiter(this, void 0, void 0, function* () {
            data.model = modelsQueue.modelNameToModelMap[groupId];
            if (!data.model.schema.options.sortEncryptedFieldsOptions.silent) {
                const noOfPendingJobs = (yield modelsQueue.client.getMetrics(100)).topMessageGroupsMessageBacklogLength;
                console.log(`mongoose-sort-encrypted-field -> handleMessage() -> noOfPendingJobs: ${noOfPendingJobs}`);
            }
            if (data.generateSortIdForAllDocuments) {
                yield generateSortIdForAllDocuments(data);
                return;
            }
            yield updateSortFieldsForDocument(data);
        });
    }
    addJob(model, data) {
        return __awaiter(this, void 0, void 0, function* () {
            const modelName = model.name;
            if (!this.modelNameToModelMap[modelName]) {
                this.modelNameToModelMap[modelName] = model;
            }
            yield this.client.send({ data, groupId: model.modelName });
        });
    }
    registerModel(model) {
        if (!this.modelNameToModelMap[model.modelName]) {
            this.modelNameToModelMap[model.modelName] = model;
        }
    }
}
function getModelsQueue(redisOptions) {
    if (!modelsQueue) {
        modelsQueue = new ModelsQueue(redisOptions);
    }
    return modelsQueue;
}
module.exports = getModelsQueue;
